/*
 * Copyright (C) 2010  Axel Kesseler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "stdafx.h"
#include "button.h"
#include "string.h"
#include "application.h"

using namespace plexdata::autoplay;

const COLORREF CButton::NORMAL_BORDER    = COLORREF( RGB( 128, 128, 128 ) );
const COLORREF CButton::HIGHLIGHT_BORDER = COLORREF( RGB( 160, 160, 160 ) );
const COLORREF CButton::FOCUSED_BORDER   = COLORREF( RGB( 192, 192, 192 ) );
const int      CButton::BORDER_MARGIN    = 4;

CButton::CButton()
	: m_hButton(NULL)
	, m_hParent(NULL)
	, m_hTooltip(NULL)
	, m_hImage(NULL)
	, m_hFont(NULL)
	, m_hCursor(NULL)
	, m_pPrevHandler(NULL)
	, m_bGotFocus(false)
	, m_bTrackLeave(false)
	, m_nAlignment(DT_CENTER)
	, m_nButtonID(0)
{
}

CButton::~CButton()
{
	::DeleteObject( this->m_hImage );
	::DeleteObject( this->m_hFont );
	::DestroyCursor( this->m_hCursor );
	::DestroyWindow( this->m_hTooltip );
}

bool CButton::Create( HWND hParent, int nButtonID, CButtonData* pButton, int x, int y, int cx, int cy )
{
	_ASSERT( hParent != NULL );
	_ASSERT( pButton != NULL );

	if ( pButton != NULL )
	{
		this->m_hParent   = hParent;
		this->m_nButtonID = nButtonID;
		this->m_sLabel    = pButton->Label();
		this->m_hImage    = pButton->LoadBitmap();
		this->m_hCursor   = ::LoadCursor
			( CApplication::GetInstance().InstanceHandle()
			, MAKEINTRESOURCE( IDC_MOUSEHAND ) 
			);

		this->m_hButton = ::CreateWindowEx
			( WS_EX_NOPARENTNOTIFY
			, _T("button")
			, _T("")
			, WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_OWNERDRAW 
			, x
			, y
			, cx
			, cy
			, this->m_hParent
			, (HMENU)this->m_nButtonID
			, CApplication::GetInstance().InstanceHandle()
			, reinterpret_cast<void*>( this )
			);

		if ( this->m_hButton != NULL )
		{
			::SetWindowLongPtr
				( this->m_hButton
				, GWLP_USERDATA
				, reinterpret_cast<INT_PTR>( this ) 
				);

			this->m_pPrevHandler = reinterpret_cast<WNDPROC>( 
				::SetWindowLongPtr
					( this->m_hButton
					, GWLP_WNDPROC
					, reinterpret_cast<INT_PTR>( CButton::EventHandler ) 
					)
				);

			// Prepare button's tooltip if needed.
			if ( !pButton->Tooltip().IsEmpty() )
			{
				// Create tooltip's window.
				this->m_hTooltip = ::CreateWindow
					( TOOLTIPS_CLASS
					, _T("")
					, TTS_ALWAYSTIP
					, CW_USEDEFAULT
					, CW_USEDEFAULT
					, CW_USEDEFAULT
					, CW_USEDEFAULT
					, this->m_hButton
					, (HMENU)NULL
					, CApplication::GetInstance().InstanceHandle()
					, NULL
					);

				_ASSERT( this->m_hTooltip != NULL );

				// Prepare tooltip's info structure.
				TOOLINFO toolInfo    = {0};
				toolInfo.cbSize      = sizeof(TOOLINFO);
				toolInfo.uFlags      = TTF_SUBCLASS;
				toolInfo.hwnd        = this->m_hButton;
				toolInfo.hinst       = CApplication::GetInstance().InstanceHandle();
				toolInfo.lpszText    = const_cast<LPTSTR>( pButton->Tooltip().String() );
				toolInfo.rect.bottom = cy;
				toolInfo.rect.right  = cx; 

				// Add this tooltip to the button.
				::SendMessage( this->m_hTooltip, TTM_ADDTOOL, (WPARAM)0, (LPARAM)&toolInfo );
			}
		}
	}

	return ( this->m_hButton != NULL );
}

bool CButton::SetFont( LPCTSTR szFont, int nSize, bool bBold, int nAlignment )
{
	if ( this->m_hButton != NULL && szFont != NULL )
	{
		this->m_nAlignment = nAlignment;

		::DeleteObject( this->m_hFont );
		this->m_hFont = NULL;

		HDC hdcButton = ::GetDC( this->m_hButton );

		// Create button text font.
		this->m_hFont = ::CreateFont
			( -MulDiv( nSize, ::GetDeviceCaps( hdcButton, LOGPIXELSY ), 72 )
			, 0
			, 0
			, 0
			, bBold ? FW_BOLD : 0
			, FALSE
			, FALSE
			, FALSE
			, DEFAULT_CHARSET
			, OUT_DEFAULT_PRECIS
			, CLIP_DEFAULT_PRECIS
			, DEFAULT_QUALITY
			, DEFAULT_PITCH
			, szFont
			);

		_ASSERT( this->m_hFont != NULL );

		::ReleaseDC( this->m_hButton, hdcButton );

		return ( this->m_hFont != NULL );
	}
	else
	{
		return false;
	}
}

LRESULT CALLBACK CButton::EventHandler( HWND hButton, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	CButton* pButton = reinterpret_cast<CButton*>( ::GetWindowLongPtr( hButton, GWLP_USERDATA ) );

	switch ( uMsg )
	{
	case WM_PAINT:

		if ( pButton != NULL )
		{
			PAINTSTRUCT sPaint = {0};
			pButton->OnPaint( ::BeginPaint( hButton, &sPaint ) );
			::EndPaint( hButton, &sPaint );
		}
		break;

	case WM_MOUSEMOVE:

		if ( pButton != NULL )
		{
			pButton->OnMouseMove( LOWORD(lParam), HIWORD(lParam) );
		}
		break;

	case WM_MOUSELEAVE:

		if ( pButton != NULL )
		{
			pButton->OnMouseLeave();
		}
		break;

	case WM_SETFOCUS:

		if ( pButton != NULL )
		{
			pButton->OnSetFocus();
		}
		break;

	case WM_KILLFOCUS:

		if ( pButton != NULL )
		{
			pButton->OnKillFocus();
		}
		break;

	case WM_KEYDOWN:

		if ( pButton != NULL )
		{
			pButton->OnKeyDown( wParam );
		}
		break;

	default:

		// Nothing else to do.
		break;
	}

	if ( pButton != NULL )
	{
		return ::CallWindowProc( pButton->m_pPrevHandler, hButton, uMsg, wParam, lParam );
	}
	else
	{
		return ::DefWindowProc( hButton, uMsg, wParam, lParam );
	}
}

void CButton::OnPaint( HDC hdcTarget )
{
	RECT     rClient    = {0};
	RECT     rText      = {0};
	RECT     rCalc      = {0};
	RECT     rImage     = {0};
	SIZE     sSize      = {0};
	HFONT    hOldFont   = NULL;
	int      nDelta     = 0;
	int      nLines     = 0;
	int      nOldBkMode = 0;
	COLORREF nOldColor  = 0;
	int      nFormat    = this->m_nAlignment | DT_WORDBREAK | DT_CALCRECT;

	// Use button text font.
	hOldFont = reinterpret_cast<HFONT>( ::SelectObject( hdcTarget, this->m_hFont ) );

	// Pre-calculate needed text box size.
	::GetClientRect( this->m_hButton, &rClient );

	rText.left   = rClient.left   + CButton::BORDER_MARGIN;
	rText.top    = rClient.top    + CButton::BORDER_MARGIN;
	rText.right  = rClient.right  - CButton::BORDER_MARGIN;
	rText.bottom = rClient.bottom - CButton::BORDER_MARGIN;

	if ( this->m_hImage != NULL )
	{
		rText.right -= rText.bottom - rText.top + CButton::BORDER_MARGIN;

		rImage.left   = rText.right + CButton::BORDER_MARGIN;
		rImage.top    = rText.top;
		rImage.right  = rClient.right - CButton::BORDER_MARGIN;
		rImage.bottom = rText.bottom;
	}

	rCalc.left   = rText.left;
	rCalc.top    = rText.top;
	rCalc.right  = rText.right;
	rCalc.bottom = rText.bottom;

	::DrawText( hdcTarget, this->m_sLabel.String(), -1, &rCalc, nFormat );

	// Calculate allowed number of lines.
	::GetTextExtentPoint32( hdcTarget, _T("W"), 1, &sSize );
	nLines = ( rText.bottom - rText.top ) / sSize.cy;

	if ( rCalc.bottom - rCalc.top > nLines * sSize.cy )
	{
		// Fit text box size to calulated line limit.
		rText.bottom = rText.top + nLines * sSize.cy;
	}
	else
	{
		// Use calculated rectangle instead.
		rText.right  = rCalc.right;
		rText.left   = rCalc.left;
		rText.top    = rCalc.top;
		rText.bottom = rCalc.bottom;
	}

	// Vertical alignment (center only) calculation.
	nDelta = ( rClient.bottom - rClient.top - rText.bottom + rText.top ) / 2;
	rText.top    += nDelta - CButton::BORDER_MARGIN;
	rText.bottom += nDelta - CButton::BORDER_MARGIN;

	// Horizontal alignment calculation.
	if ( this->m_nAlignment == DT_CENTER )
	{
		if ( this->m_hImage != NULL )
		{
			nDelta = ( rImage.left - rClient.left - rText.right + rText.left ) / 2;
		}
		else
		{
			nDelta = ( rClient.right - rClient.left - rText.right + rText.left ) / 2;
		}
		rText.left  += nDelta - CButton::BORDER_MARGIN;
		rText.right += nDelta - CButton::BORDER_MARGIN;
	}
	else if ( this->m_nAlignment == DT_RIGHT )
	{
		rText.right = rClient.right - CButton::BORDER_MARGIN;
		if ( this->m_hImage != NULL )
		{
			rText.right = rImage.left - CButton::BORDER_MARGIN;
		}
		rText.left = rText.right - ( rCalc.right - rCalc.left );
	}

	// Use transparent drawing background mode.
	nOldBkMode = ::SetBkMode( hdcTarget, TRANSPARENT ); 

	#ifdef _DEBUG
	::FrameRect( hdcTarget, &rText, (HBRUSH)::GetStockObject( BLACK_BRUSH ) );
	#endif

	// Set text color to be used.
	nOldColor = ::SetTextColor( hdcTarget, CApplication::GetInstance().Config().FontColor() );

	// Now we are ready to paint the welcome text.
	nFormat &= ~DT_CALCRECT;
	::DrawText( hdcTarget, this->m_sLabel.String(), -1, &rText, nFormat );

	// Restore original text color! Do it before painting the image.
	::SetTextColor( hdcTarget, nOldColor );

	if ( this->m_hImage != NULL )
	{
		BITMAP bmpInfo = {0};

		::GetObject( this->m_hImage, sizeof(BITMAP), &bmpInfo );

		int cx = bmpInfo.bmWidth;
		int cy = bmpInfo.bmHeight;

		if ( cx > rImage.right - rImage.left )
		{
			cx = rImage.right - rImage.left;
		}

		if ( cy > rImage.bottom - rImage.top )
		{
			cy = rImage.bottom - rImage.top;
		}

		this->DrawImage( hdcTarget, rImage.left, rImage.top, cx, cy );

		#ifdef _DEBUG
		::FrameRect( hdcTarget, &rImage, (HBRUSH)::GetStockObject( BLACK_BRUSH ) );
		#endif
	}

	// Restore original background drawing mode.
	::SetBkMode( hdcTarget, nOldBkMode ); 

	// Restore original font.
	::SelectObject( hdcTarget, hOldFont );

	// Draw button's border.
	this->DrawBorder();
}

void CButton::OnMouseMove( int xPos, int yPos )
{
	::SetCursor( this->m_hCursor );

	if ( !this->m_bTrackLeave )
	{
		TRACKMOUSEEVENT trackEvents = {0};

		// Lock mouse tracking.
		this->m_bTrackLeave = true;

		trackEvents.cbSize    = sizeof(TRACKMOUSEEVENT);
		trackEvents.dwFlags   = TME_LEAVE;
		trackEvents.hwndTrack = this->m_hButton;

		::TrackMouseEvent( &trackEvents );

		// Draw button's border.
		this->DrawBorder();
	}
}

void CButton::OnMouseLeave()
{
	// Unlock mouse tracking.
	this->m_bTrackLeave = false;

	// Draw button's border.
	this->DrawBorder();
}

void CButton::OnSetFocus()
{
	// Lock keyboard focus.
	this->m_bGotFocus = true;

	// Draw button's border.
	this->DrawBorder();
}

void CButton::OnKillFocus()
{
	// Unlock keyboard focus.
	this->m_bGotFocus = false;

	// Draw button's border.
	this->DrawBorder();
}

void CButton::OnKeyDown( unsigned int nVirtualKey )
{
	switch ( nVirtualKey )
	{
	case VK_UP:

		// Move input focus to previous control.
		::SetFocus( ::GetNextDlgTabItem( this->m_hParent, this->m_hButton, true ) );
		break;

	case VK_DOWN:

		// Move input focus to next control.
		::SetFocus( ::GetNextDlgTabItem( this->m_hParent, this->m_hButton, false ) );
		break;

	case VK_TAB:

		// Move input focus to next sibling control using current shift key state.
		::SetFocus( 
			::GetNextDlgTabItem
				( this->m_hParent
				, this->m_hButton
				, ( ::GetKeyState( VK_SHIFT ) & 0x8000 ) ? true : false
			));
		break;

	case VK_RETURN:
		
		// Send button click to parent window (same as VK_SPACE event).
		::SendMessage
			( this->m_hParent
			, WM_COMMAND
			, MAKEWPARAM( this->m_nButtonID, BN_CLICKED )
			, (LPARAM)this->m_hButton
			);
		break;
	}
}

void CButton::DrawBorder()
{
	RECT     rClient   = {0};
	HDC      hdcButton = NULL;
	COLORREF crBorder  = 0;

	if ( this->m_bGotFocus ) // Current button has the input focus.
	{
		crBorder = CButton::FOCUSED_BORDER;
	}
	else if ( this->m_bTrackLeave ) // Wait for mouse leaves current button
	{
		crBorder = CButton::HIGHLIGHT_BORDER;
	}
	else
	{
		crBorder = CButton::NORMAL_BORDER;
	}

	::GetClientRect( this->m_hButton, &rClient );

	hdcButton = ::GetDC( this->m_hButton );

	HPEN hPen = ::CreatePen( PS_SOLID, 2, crBorder );
	HPEN hOld = (HPEN)::SelectObject( hdcButton, hPen );

	::MoveToEx( hdcButton, rClient.left, rClient.top, NULL );

	::LineTo( hdcButton, rClient.right, rClient.top);
	::LineTo( hdcButton, rClient.right, rClient.bottom);
	::LineTo( hdcButton, rClient.left,  rClient.bottom);
	::LineTo( hdcButton, rClient.left,  rClient.top - 1); // BUGFIX: Fixes a missing pixel on that location!

	::SelectObject( hdcButton, hOld );
	::DeleteObject( hPen );
	::ReleaseDC( this->m_hButton, hdcButton );
}

void CButton::DrawImage( HDC hdcTarget, int x, int y, int cx, int cy )
{
	HDC      hdcImage      = NULL;
	HDC      hdcTrans      = NULL;
	HBITMAP  hbmTrans      = NULL;
	HBITMAP  hbmOldImage   = NULL;
	HBITMAP  hbmOldTrans   = NULL;
	COLORREF crTransparent = 0;

	// Create input bitmap.
	hdcImage      = ::CreateCompatibleDC( hdcTarget );
	hbmOldImage   = reinterpret_cast<HBITMAP>( ::SelectObject( hdcImage, this->m_hImage ) );
	crTransparent = ::GetPixel( hdcImage, 0, 0 ); // Get transparent color pixel.

	// Create mask bitmap.
	hdcTrans    = ::CreateCompatibleDC( hdcTarget );
	hbmTrans    = ::CreateBitmap( cx, cy, 1, 1, NULL );
	hbmOldTrans = reinterpret_cast<HBITMAP>( ::SelectObject( hdcTrans, hbmTrans ) );

	// Build the mask based on transparent color.
	::SetBkColor( hdcImage, crTransparent );
	::BitBlt( hdcTrans, 0, 0, cx, cy, hdcImage, 0, 0, SRCCOPY );

	// Do the transparent blit.
	::BitBlt( hdcTarget, x, y, cx, cy, hdcImage, 0, 0, SRCINVERT );
	::BitBlt( hdcTarget, x, y, cx, cy, hdcTrans, 0, 0, SRCAND );
	::BitBlt( hdcTarget, x, y, cx, cy, hdcImage, 0, 0, SRCINVERT );

	// Restore settings
	::SelectObject( hdcImage, hbmOldImage );
	::SelectObject( hdcTrans, hbmOldTrans );
	::DeleteDC( hdcImage );
	::DeleteDC( hdcTrans );
}
