/*
 * Copyright (C) 2010  Axel Kesseler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "stdafx.h"
#include <shellapi.h>
#include "dialog.h"
#include "string.h"
#include "splitpath.h"
#include "version.h"
#include "application.h"

using namespace plexdata::autoplay;

const int CDialog::WELCOME_LINES  = 4;
const int CDialog::WELCOME_HEADER = 15;
const int CDialog::WELCOME_MARGIN = 25;
const int CDialog::BUTTON_MARGIN  = 7;
const int CDialog::BUTTON_ID_BASE = 4711;

CDialog::CDialog()
	: m_hDialog(NULL)
	, m_hBkBrush(NULL)
	, m_hFont(NULL)
	, m_pButtons(NULL)
	, m_nFormat(DT_TOP|DT_CENTER|DT_WORDBREAK)
	, m_bExecuting(false)
	, m_bClosing(false)
{
}

CDialog::~CDialog()
{
	::DeleteObject( this->m_hBkBrush );
	::DeleteObject( this->m_hFont );
	delete[] this->m_pButtons;
}

bool CDialog::Initialize()
{
	bool success = false;

	this->m_hBkBrush = ::CreateSolidBrush( COLORREF( RGB( 255, 255, 255 ) ) );

	this->m_hDialog = ::CreateDialogParam
		( CApplication::GetInstance().InstanceHandle()
		, MAKEINTRESOURCE( IDD_AUTOPLAY )
		, NULL
		, CDialog::DialogProc
		, reinterpret_cast<LPARAM>( this )
		);

	if ( this->m_hDialog != NULL )
	{
		// Adjust dialog box's border depending on given configuration.
		this->AdjustBorderStyle( this->m_hDialog );

		// Change window class icon which appear in the tast switch.
		::SetClassLong
			( this->m_hDialog
			, GCL_HICON
			, (LONG)CApplication::GetInstance().Config().Icon() 
			);

		// Set dialog window's icon.
		::SendMessage
			( this->m_hDialog
			, WM_SETICON
			, (WPARAM)ICON_SMALL
			, (LPARAM)CApplication::GetInstance().Config().Icon()
			);

		// Set dialog window's caption.
		::SendMessage
			( this->m_hDialog
			, WM_SETTEXT
			, (WPARAM)0
			, (LPARAM)CApplication::GetInstance().Config().Title()
			);

		// Insert About box menu item within the system menu.
		HMENU hMenu = ::GetSystemMenu( this->m_hDialog, FALSE );
		if ( hMenu != NULL )
		{
			int nCount = ::GetMenuItemCount( hMenu );
			for (int index = 0; index < nCount; index++)
			{
				if ( ::GetMenuItemID( hMenu, index ) == SC_CLOSE )
				{
					plexdata::string::CString sAbout;

					// Add a separator in front of new menu items.
					::InsertMenu( hMenu, index, MF_SEPARATOR|MF_BYPOSITION, 0, _T("") );

					// Get language depending Help menu item string.
					::LoadString
						( CApplication::GetInstance().InstanceHandle()
						, IDS_HELP
						, sAbout.CreateBuffer( 32 )
						, 31 // Ensure terminating '\0'!
						);
					
					// Add Help menu item to the system menu.
					sAbout.CopyBuffer();
					::InsertMenu( hMenu, index, MF_BYPOSITION|MF_STRING, IDM_HELP, sAbout.String() );

					// Get language depending About menu item string.
					::LoadString
						( CApplication::GetInstance().InstanceHandle()
						, IDS_ABOUT
						, sAbout.CreateBuffer( 32 )
						, 31 // Ensure terminating '\0'!
						);

					// Add About menu item to the system menu.
					sAbout.CopyBuffer();
					::InsertMenu( hMenu, index, MF_BYPOSITION|MF_STRING, IDM_ABOUT, sAbout.String() );
				}
			}
		}

		success = true;
	}
	return success;
}

int CDialog::Show
	( int nCmdShow 
	)
{
	int result = 0;

	if ( this->m_hDialog )
	{
		::ShowWindow( this->m_hDialog,  nCmdShow );
		::InvalidateRect( this->m_hDialog, NULL, TRUE );

		MSG msg = {0};
		while ( ::GetMessage( &msg, NULL, 0, 0 ) )
		{
			if ( !::TranslateAccelerator( msg.hwnd, NULL, &msg ) )
			{
				::TranslateMessage( &msg );
				::DispatchMessage( &msg );
			}
		}

		result = (int)msg.wParam;
	}

	return result;
}

INT_PTR CALLBACK CDialog::DialogProc
	( HWND   hDialog
	, UINT   uMsg
	, WPARAM wParam
	, LPARAM lParam 
	)
{
	INT_PTR         nResult = FALSE; // Assume message couldn't be handled.
	static CDialog* pDialog = NULL;

	switch ( uMsg )
	{
	case WM_INITDIALOG:

		pDialog = reinterpret_cast<CDialog*>( lParam );

		_ASSERT( pDialog != NULL );

		if ( pDialog != NULL )
		{
			pDialog->OnInitDialog( hDialog );
			nResult = TRUE; // WM_INITDIALOG should return TRUE (see MSDN).
		}
		break;

	case WM_CTLCOLORDLG:
	case WM_CTLCOLORBTN:

		if ( pDialog != NULL )
		{
			nResult = reinterpret_cast<LRESULT>( pDialog->m_hBkBrush ); // A brush has to be returned instead!
		}
		break;

	case WM_PAINT:

		if ( pDialog != NULL )
		{
			PAINTSTRUCT sPaint= {0};
			pDialog->OnPaint( ::BeginPaint( hDialog, &sPaint ) );
			::EndPaint( hDialog, &sPaint );
			nResult = TRUE;  // Message was handled.
		}
		break;

	case WM_HELP:

		CApplication::GetInstance().ShowHelp( hDialog );
		break;

	case WM_CLOSE:
		
		if ( pDialog != NULL )
		{
			pDialog->OnClose();
			nResult = TRUE;  // Message was handled.
		}
		break;

	case WM_COMMAND:
		
		if ( HIWORD(wParam) == BN_CLICKED )
		{
			if ( pDialog != NULL )
			{
				pDialog->OnButtonClick( LOWORD(wParam), reinterpret_cast<HWND>( lParam ) );
				nResult = TRUE;  // Message was handled.
			}
		}
		break;

	case WM_SYSCOMMAND:
		
		if ( pDialog != NULL )
		{
			nResult = ( pDialog->OnSysCommand( wParam, lParam ) ? TRUE : FALSE );
		}
		break;

	case WM_INITMENUPOPUP:

		if ( pDialog != NULL )
		{
			pDialog->OnInitMenuPopup
				( reinterpret_cast<HMENU>(wParam)
				, LOWORD(lParam)
				, HIWORD(lParam) ? true : false
				);
			nResult = TRUE;  // Message was handled.
		}
		break;

	default:

		nResult = FALSE;  // Message was not handled!
		break;
	}

	return nResult;
}

INT_PTR CALLBACK CDialog::AboutHandler
	( HWND   hAbout
	, UINT   uMsg
	, WPARAM wParam
	, LPARAM lParam 
	)
{
	switch ( uMsg )
	{
	case WM_INITDIALOG:
		{
			plexdata::version::CVersionInfo version;
			plexdata::string::CString       helper;

			version.Initialize( CApplication::GetInstance().ModuleFile() );

			helper.Format( _T("%s, Version %s"), version.ProductName(), version.ProductVersion() );
			SetDlgItemText( hAbout, IDC_PRODUCT, helper.String() );

			helper.Format( _T("%s %s"), version.LegalCopyright(), version.CompanyName() );
			SetDlgItemText( hAbout, IDC_COPYRIGHT, helper.String() );

			SetDlgItemText( hAbout, IDC_DESCRIPTION, version.FileDescription() );

			SetDlgItemText( hAbout, IDC_COMMENTS, version.Comments() );
		}
		return TRUE; // WM_INITDIALOG should return TRUE (see MSDN).

	case WM_CTLCOLORDLG:
	case WM_CTLCOLORSTATIC:

		return reinterpret_cast<INT_PTR>( ::GetStockObject( WHITE_BRUSH ) );

	case WM_COMMAND:
		
		if ( LOWORD(wParam) == IDOK )
		{
			::EndDialog( hAbout, IDOK );
			return TRUE;
		}
		else if ( LOWORD(wParam) == IDHELP )
		{
			CWaitCursor wait;
			if ( CApplication::GetInstance().ShowHelp( ::GetParent( hAbout ) ) )
			{
				::EndDialog( hAbout, IDOK );
			}
			return TRUE;
		}
		break;

	case WM_CLOSE:

		::EndDialog( hAbout, IDOK );
		return TRUE;

	default:

		// Nothing else to do.
		break;
	}

	return FALSE;
}

void CDialog::OnInitDialog
	( HWND hDialog 
	)
{
	HDC  hdcDialog = ::GetDC( hDialog ); // Get dialog's device context.
	RECT rDesktop  = {0};
	RECT rDialog   = {0};

	//
	// Initialize background.
	//
	if ( CApplication::GetInstance().Config().Background() != NULL )
	{
		HDC     hdcSource = ::CreateCompatibleDC( hdcDialog );
		HDC     hdcTarget = ::CreateCompatibleDC( hdcDialog );
		HBITMAP hbmSource = CApplication::GetInstance().Config().Background();
		HBITMAP hbmTarget = NULL;
		BITMAP  bmpSource = {0};
		int     nWidth    = 0;
		int     nHeight   = 0;
		bool    stretch   = false;
		bool    resize    = false;

		// Calculate stretch/resize parameters.
		::GetObject( hbmSource, sizeof(BITMAP), &bmpSource );

		::GetClientRect( hDialog, &rDialog );

		nWidth  = rDialog.right - rDialog.left;
		nHeight = rDialog.bottom - rDialog.top;

		if ( nWidth > bmpSource.bmWidth )
		{			
			stretch = true;
		}
		else
		{
			nWidth = bmpSource.bmWidth;
			resize = true;
		}

		if ( nHeight > bmpSource.bmHeight )
		{
			stretch = true;
		}
		else
		{
			nHeight = bmpSource.bmHeight;
			resize = true;
		}

		// Resize AND center dialog window.
		if ( resize )
		{
			::GetClientRect( ::GetDesktopWindow(), &rDesktop );

			::SetWindowPos
				( hDialog
				, NULL
				, ( ( rDesktop.right - rDesktop.left ) - nWidth ) / 2
				, ( ( rDesktop.bottom - rDesktop.top ) - nHeight )/ 2
				, nWidth
				, nHeight
				, SWP_NOZORDER 
				);
		}

		// Prepare source device context.
		::SelectObject( hdcSource, hbmSource );

		// Prepare target device context.
		hbmTarget = ::CreateCompatibleBitmap( hdcDialog, nWidth, nHeight );
		::SelectObject( hdcTarget, hbmTarget );
		::BitBlt( hdcTarget, 0, 0, nWidth, nHeight, hdcSource, 0, 0, WHITENESS ); 

		if ( stretch )
		{
			::StretchBlt( hdcTarget, 0, 0, nWidth, nHeight, hdcSource, 0, 0, bmpSource.bmWidth, bmpSource.bmHeight, SRCCOPY ); 
		}
		else
		{
			::BitBlt( hdcTarget, 0, 0, nWidth, nHeight, hdcSource, 0, 0, SRCCOPY ); 
		}

		// Use new created image as background brush.
		::DeleteObject( this->m_hBkBrush );
		this->m_hBkBrush = ::CreatePatternBrush( hbmTarget );

		// Clean up...
		::DeleteDC( hdcSource );
		::DeleteDC( hdcTarget );
		::DeleteObject( hbmTarget );
	}

	//
	// Initialize welcome text.
	//
	plexdata::string::CString sWelcome = CApplication::GetInstance().Config().Welcome();
	if ( sWelcome != _T("") )
	{
		// Create welcome text font.
		this->m_hFont = ::CreateFont
			( -MulDiv( CApplication::GetInstance().Config().FontSize(), ::GetDeviceCaps( hdcDialog, LOGPIXELSY ), 72 )
			, 0
			, 0
			, 0
			, CApplication::GetInstance().Config().FontBold() ? FW_BOLD : 0
			, FALSE
			, FALSE
			, FALSE
			, DEFAULT_CHARSET
			, OUT_DEFAULT_PRECIS
			, CLIP_DEFAULT_PRECIS
			, DEFAULT_QUALITY
			, DEFAULT_PITCH
			, CApplication::GetInstance().Config().FontName()
			);

		_ASSERT( this->m_hFont != NULL );
	}

	//
	// Button calculation!
	//
	int   nOffset  = 0;
	RECT  hdrRect  = this->CalcWelcomeRectangle( hDialog, this->m_hFont, this->m_nFormat );
	SIZE  dlgSize  = {0,0};
	RECT  dlgRect  = {0,0,0,0};
	SIZE  btnSize  = CApplication::GetInstance().Config().ButtonSize();
	int   btnCount = CApplication::GetInstance().Config().GetButtonCount();
	RECT* btnRect  = new RECT[btnCount];

	if ( btnRect != NULL )
	{
		// Calculate button area y-offset depending on header size.
		if ( hdrRect.bottom > 0 )
		{
			nOffset = hdrRect.bottom + CDialog::WELCOME_HEADER - CDialog::BUTTON_MARGIN;
		}
		else
		{
			nOffset = hdrRect.bottom;
		}

		// Calculate the rectangle to display the buttons within.
		::GetClientRect( hDialog, &rDialog ); // Important, don't delete!

		dlgRect.left   = rDialog.left   + CDialog::BUTTON_MARGIN;
		dlgRect.top    = rDialog.top    + CDialog::BUTTON_MARGIN + nOffset;
		dlgRect.right  = rDialog.right  - CDialog::BUTTON_MARGIN;
		dlgRect.bottom = rDialog.bottom - CDialog::BUTTON_MARGIN;

		dlgSize.cx = dlgRect.right - dlgRect.left;
		dlgSize.cy = dlgRect.bottom - dlgRect.top;

		for ( int nIndex=0; nIndex < btnCount; nIndex++ )
		{
			int xButton  = 0;
			int yButton  = 0;
			int cyButton = btnSize.cy + CDialog::BUTTON_MARGIN;

			// Current button's horitontal position
			if ( CApplication::GetInstance().Config().HorizontalAlignment() == CConfiguration::ALIGNMENT_LEFT )
			{
				xButton = dlgRect.left;
			}
			else if ( CApplication::GetInstance().Config().HorizontalAlignment() == CConfiguration::ALIGNMENT_CENTER )
			{
				xButton =  ( dlgSize.cx - btnSize.cx ) / 2;
			}
			else if ( CApplication::GetInstance().Config().HorizontalAlignment() == CConfiguration::ALIGNMENT_RIGTH )
			{
				xButton =  dlgRect.right - btnSize.cx;
			}
			else
			{
				_ASSERT( false );
			}

			// Current button's vertircal position
			if ( CApplication::GetInstance().Config().VerticalAlignment() == CConfiguration::ALIGNMENT_TOP )
			{
				yButton = dlgRect.top + cyButton * nIndex;
			}
			else if ( CApplication::GetInstance().Config().VerticalAlignment() == CConfiguration::ALIGNMENT_CENTER )
			{
				int nScreen = dlgRect.bottom - dlgRect.top;
				int nNeeded = cyButton * btnCount - CDialog::BUTTON_MARGIN;

				yButton = ( dlgRect.top + ( nScreen - nNeeded ) / 2 ) + cyButton * nIndex;
			}
			else if ( CApplication::GetInstance().Config().VerticalAlignment() == CConfiguration::ALIGNMENT_BOTTOM )
			{
				yButton = ( dlgRect.bottom + CDialog::BUTTON_MARGIN ) - cyButton * ( btnCount - nIndex );
			}
			else
			{
				_ASSERT( false );
			}
			
			// Save current button's coordinates.
			btnRect[nIndex].left   = xButton;
			btnRect[nIndex].top    = yButton;
			btnRect[nIndex].right  = xButton + btnSize.cx;
			btnRect[nIndex].bottom = yButton + btnSize.cy;
		}

		// Create dialog's button list.
		delete[] this->m_pButtons;
		this->m_pButtons = new CButton[btnCount];
		if ( this->m_pButtons != NULL )
		{
			for ( int nIndex=0; nIndex < btnCount; nIndex++ )
			{
				// Check dialog height limit violation.
				if ( btnRect[nIndex].top + btnSize.cy <= dlgRect.bottom      && 
					 dlgRect.top                      <= btnRect[nIndex].top && 
					 nOffset                          <  btnRect[nIndex].top )
				{
					// Button can be displayed!
					this->m_pButtons[nIndex].Create
						( hDialog
						, CDialog::BUTTON_ID_BASE + nIndex
						, CApplication::GetInstance().Config().GetButtonEntry( nIndex )
						, btnRect[nIndex].left
						, btnRect[nIndex].top
						, btnSize.cx
						, btnSize.cy
						);

					this->m_pButtons[nIndex].SetFont
						( CApplication::GetInstance().Config().FontName()
						, CApplication::GetInstance().Config().FontSize()
						, CApplication::GetInstance().Config().FontBold()
						, CApplication::GetInstance().Config().GetButtonEntry( nIndex )->Alignment()
						);
				}
			}
		}

		delete[] btnRect;
	}

	::ReleaseDC( hDialog, hdcDialog );
}

void CDialog::OnPaint
	( HDC hdcTarget 
	)
{
	plexdata::string::CString sWelcome = CApplication::GetInstance().Config().Welcome();
	RECT     rWelcome   = {0};
	HFONT    hOldFont   = NULL;
	int      nOldBkMode = 0;
	COLORREF nOldColor  = 0;

	if ( CApplication::GetInstance().Config().GetButtonCount() == 0 )
	{
		RECT rFrame = {0};
		::GetClientRect( this->m_hDialog, &rFrame );
		rFrame.left   += CDialog::BUTTON_MARGIN;
		rFrame.top    += CDialog::BUTTON_MARGIN * 5;
		rFrame.right  -= CDialog::BUTTON_MARGIN;
		rFrame.bottom -= CDialog::BUTTON_MARGIN;

		plexdata::string::CString sHelp;
		::LoadString
			( CApplication::GetInstance().InstanceHandle()
			, IDS_SHORTCUTS
			, sHelp.CreateBuffer( 256 )
			, 255 // Ensure terminating '\0'!
			);
		sHelp.CopyBuffer();

		// Use pre-configured font or system default font if not exist.
		hOldFont = reinterpret_cast<HFONT>( ::SelectObject( hdcTarget, this->m_hFont ) );

		// Use transparent drawing background mode.
		nOldBkMode = ::SetBkMode( hdcTarget, TRANSPARENT ); 

		// Set text color to be used.
		nOldColor = ::SetTextColor( hdcTarget, CApplication::GetInstance().Config().FontColor() );

		// Now we are ready to paint the help text.
		::DrawText( hdcTarget, sHelp.String(), -1, &rFrame, this->m_nFormat );

		// Restore original text color! Do it before painting the image.
		::SetTextColor( hdcTarget, nOldColor );

		// Restore original background drawing mode.
		::SetBkMode( hdcTarget, nOldBkMode ); 

		// Restore original font.
		::SelectObject( hdcTarget, hOldFont );
	}
	else if ( sWelcome != _T("") )
	{
		// Calculate Welcome text area.
		rWelcome = this->CalcWelcomeRectangle( this->m_hDialog, this->m_hFont, this->m_nFormat );

		// Use welcome text font.
		hOldFont = reinterpret_cast<HFONT>( ::SelectObject( hdcTarget, this->m_hFont ) );

		// Use transparent drawing background mode.
		nOldBkMode = ::SetBkMode( hdcTarget, TRANSPARENT ); 

		#ifdef _DEBUG
		::FrameRect( hdcTarget, &rWelcome, (HBRUSH)::GetStockObject( BLACK_BRUSH ) );
		#endif

		// Set text color to be used.
		nOldColor = ::SetTextColor( hdcTarget, CApplication::GetInstance().Config().FontColor() );

		// Now we are ready to paint the welcome text.
		::DrawText( hdcTarget, sWelcome.String(), -1, &rWelcome, this->m_nFormat );

		// Restore original text color! Do it before painting the image.
		::SetTextColor( hdcTarget, nOldColor );

		// Restore original background drawing mode.
		::SetBkMode( hdcTarget, nOldBkMode ); 

		// Restore original font.
		::SelectObject( hdcTarget, hOldFont );
	}

	#ifdef _DEBUG
	RECT frame = {0};
	::GetClientRect( this->m_hDialog, &frame );
	frame.left   += CDialog::BUTTON_MARGIN;
	frame.top    += CDialog::BUTTON_MARGIN + ( rWelcome.bottom > 0 ) ? ( rWelcome.bottom + CDialog::WELCOME_HEADER - CDialog::BUTTON_MARGIN ) : rWelcome.bottom;
	frame.right  -= CDialog::BUTTON_MARGIN;
	frame.bottom -= CDialog::BUTTON_MARGIN;
	::FrameRect( hdcTarget, &frame, (HBRUSH)::GetStockObject( BLACK_BRUSH ) );
	#endif
}

void CDialog::OnClose()
{
	this->m_bClosing   = true;  // Avoid other actions during close action!
	this->m_bExecuting = false; // Unlock waiting!
	
	::ShowWindow( this->m_hDialog, SW_HIDE );
	::CloseWindow( this->m_hDialog );
	::DestroyWindow( this->m_hDialog );
	::PostQuitMessage( IDCLOSE );
}

void CDialog::OnButtonClick
	( int  nButtonID
	, HWND hButton 
	)
{
	CWaitCursor wait;
	CButtonData* pBtnData = CApplication::GetInstance().Config().GetButtonEntry( nButtonID - CDialog::BUTTON_ID_BASE );

	if ( pBtnData != NULL )
	{
		if ( this->m_bExecuting ) // Do this only if waiting for a process is active!
		{
			// Suppress all actions except quit action!
			if ( pBtnData->IsQuitAction() )
			{
				this->ExecuteButton( pBtnData );
			}
		}
		else
		{
			this->ExecuteButton( pBtnData );
		}
	}
}

bool CDialog::OnSysCommand
	( int nCommandID
	, int nAdditional 
	)
{
	bool handled = false; // Assume this message should be handled by default dialog procedure.

	if ( this->m_bExecuting ) // Do this only if waiting for a process is active!
	{
		// Catch some of the system menu messages!
		switch ( nCommandID & 0xFFF0 ) // The four low-order bits of wParam are used by the system, see MSDN!
		{
		case SC_MOUSEMENU:
		case SC_KEYMENU:
			handled = true; // Avoid showing of system menu if execution is currently active!
			break;
		default:
			break;          // Pass through all other system menu message (including SC_CLOSE)!
		}
	}
	else 
	{
		switch ( nCommandID )
		{
		case IDM_ABOUT:

			::DialogBox
				( CApplication::GetInstance().InstanceHandle()
				, MAKEINTRESOURCE( IDD_ABOUTBOX )
				, this->m_hDialog
				, CDialog::AboutHandler 
				);
			handled = true;  // Message handled.
			break;

		case IDM_HELP:

			::SendMessage
				( this->m_hDialog
				, WM_HELP
				, (WPARAM)NULL
				, (LPARAM)NULL // LPHELPINFO is not needed!
				);
			handled = true;  // Message handled.
			break;
		}
	}
	return handled;
}

void CDialog::OnInitMenuPopup
	( HMENU        hMenu
	, unsigned int nIndex
	, bool         bSystemMenu 
	)
{
	if ( hMenu != NULL )
	{
		if ( bSystemMenu )
		{
			struct { int nID; int nValue[2]; } defaults[] = {
				{ SC_RESTORE,  {MF_BYCOMMAND|MF_DISABLED|MF_GRAYED, MF_BYCOMMAND|MF_ENABLED            }},
				{ SC_MOVE,     {MF_BYCOMMAND|MF_ENABLED,            MF_BYCOMMAND|MF_DISABLED|MF_GRAYED }},
				{ SC_SIZE,     {MF_BYCOMMAND|MF_DISABLED|MF_GRAYED, MF_BYCOMMAND|MF_DISABLED|MF_GRAYED }},
				{ SC_MINIMIZE, {MF_BYCOMMAND|MF_ENABLED,            MF_BYCOMMAND|MF_DISABLED|MF_GRAYED }},
				{ SC_MAXIMIZE, {MF_BYCOMMAND|MF_DISABLED|MF_GRAYED, MF_BYCOMMAND|MF_DISABLED|MF_GRAYED }},
				{ SC_CLOSE,    {MF_BYCOMMAND|MF_ENABLED,            MF_BYCOMMAND|MF_ENABLED            }}
			};

			int column = ( !::IsIconic( this->m_hDialog ) ) ? 0 : 1;
			int count  = sizeof(defaults)/sizeof(defaults[0]);

			// REMARK: An adaption of the dialog's system menu is needed 
			//         only if used border style is set to tool window!
			if ( CConfiguration::BORDER_STYLE_TOOL == CApplication::GetInstance().Config().GetBorderStyle() )
			{
				for (int index=0; index < count; index++ )
				{
					::EnableMenuItem( hMenu, defaults[index].nID, defaults[index].nValue[column] );
				}
			}
		}
	}
}

void CDialog::AdjustBorderStyle
	( HWND hDialog 
	)
{
	int nBorder = CApplication::GetInstance().Config().GetBorderStyle();

	if ( NULL != hDialog && CConfiguration::BORDER_STYLE_DEFAULT != nBorder )
	{
		LONG_PTR nStyle   = ::GetWindowLongPtr( hDialog, GWL_STYLE );
		LONG_PTR nExStyle = ::GetWindowLongPtr( hDialog, GWL_EXSTYLE );

		if ( CConfiguration::BORDER_STYLE_TOOL == nBorder )
		{
			nExStyle |= WS_EX_TOOLWINDOW;
		}
		else if ( CConfiguration::BORDER_STYLE_THICK == nBorder )
		{
			nStyle &= ~WS_DLGFRAME;
		}
		else if ( CConfiguration::BORDER_STYLE_THIN == nBorder )
		{
			nStyle   &= ~WS_DLGFRAME;
			nExStyle &= ~WS_EX_DLGMODALFRAME;
		}
		else if ( CConfiguration::BORDER_STYLE_NONE == nBorder )
		{
			nStyle   &= ~(WS_DLGFRAME | WS_BORDER);
			nExStyle &= ~WS_EX_DLGMODALFRAME;
		}

		::SetWindowLongPtr( hDialog, GWL_STYLE,   nStyle   );
		::SetWindowLongPtr( hDialog, GWL_EXSTYLE, nExStyle );
		::SetWindowPos( hDialog, NULL, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER|SWP_FRAMECHANGED );
	}
}

RECT CDialog::CalcWelcomeRectangle
	( HWND  hDialog
	, HFONT hFont
	, int   nFormat
	)
{
	plexdata::string::CString sWelcome = CApplication::GetInstance().Config().Welcome();
	HDC   hdcDialog = NULL;
	HFONT hOldFont  = NULL;
	RECT  rDialog   = {0};
	RECT  rResult   = {0};
	int   nMaxSize  = 0;
	int   nDelta    = 0;

	if ( hDialog != NULL && hFont != NULL && ( sWelcome != _T("") ) )
	{
		// Get dialog's client area.
		::GetClientRect( hDialog, &rDialog );

		// Get dialog's device context.
		hdcDialog = ::GetDC( hDialog );

		// Use welcome text font.
		hOldFont = reinterpret_cast<HFONT>( ::SelectObject( hdcDialog, hFont ) );

		// Calculate maximum height of welcome text box.
		SIZE size = {0};
		::GetTextExtentPoint32( hdcDialog, _T("W"), 1, &size );
		nMaxSize = size.cy * WELCOME_LINES;

		_ASSERT( nMaxSize > 0 );

		// Pre-calculate needed welcome text box.
		rResult.left   = rDialog.left  + WELCOME_MARGIN;
		rResult.right  = rDialog.right - WELCOME_MARGIN; 
		rResult.top    = rDialog.top   + WELCOME_HEADER; 
		rResult.bottom = rDialog.top   + nMaxSize; 

		// Calculate real welcome text box.
		nFormat |= DT_CALCRECT;
		::DrawText( hdcDialog, sWelcome.String(), -1, &rResult, nFormat );

		// Limit welcome text box if it exceeds the maximum size.
		if ( rResult.bottom - rResult.top > nMaxSize )
		{
			rResult.bottom = rResult.top + nMaxSize;
		}

		// Center welcome text box on the dialog window!
		nDelta = ( rDialog.right - rDialog.left - rResult.right + rResult.left ) / 2;

		rResult.left  += nDelta - WELCOME_MARGIN;
		rResult.right += nDelta - WELCOME_MARGIN;

		// Clean up.
		::ReleaseDC( hDialog, hdcDialog );
	}

	return rResult;
}

void CDialog::ExecuteButton
	( CButtonData* pBtnData 
	)
{
	static int nRecursions = 0;

	_ASSERT( pBtnData != NULL );

	// Avoid execution during close action!
	if ( this->m_bClosing )
	{
		return; // Breaks the recursion too!
	}

	// Perform quit action.
	if ( pBtnData->IsQuitAction() )
	{
		::SendMessage( this->m_hDialog, WM_CLOSE, (WPARAM)0, (LPARAM)0 );
	}
	// Perform browse action.
	else if ( pBtnData->IsBrowseAction() )
	{
		::ShellExecute
			( NULL
			, _T("open")
			, CApplication::GetInstance().ModulePath()
			, NULL
			, NULL
			, SW_SHOWNORMAL 
			);
	}
	// Perform all other actions.
	else
	{
		// Build full qualified executable file name.
		plexdata::string::CString sExecute;
		sExecute.Format( _T("\"%s%s\""), CApplication::GetInstance().ModulePath(), pBtnData->Execute().String() );

		// Build working directory based on executable's file name.
		plexdata::string::CString sDirectory;
		sDirectory = CApplication::GetInstance().ModulePath();
		sDirectory += pBtnData->Execute();

		plexdata::string::CSplitPath splitHelper( sDirectory.String() );
		splitHelper.Filename( NULL );
		splitHelper.Extension( NULL );

		plexdata::string::CString sCommandLine;
		sCommandLine.Format( _T("%s %s"), sExecute.String(), pBtnData->Parameter().String() );

		if ( !this->LaunchProcess( sCommandLine.String(), pBtnData->HasAfterwards() ) )
		{
			// Try to start process via windows shell. This makes 
			// sense if configured executable is not an EXE file.
			::ShellExecute
				( NULL
				, _T("open")
				, sExecute.String()
				, pBtnData->Parameter().String()
				, splitHelper.Build()
				, SW_SHOWNORMAL 
				);
		}
	}

	if ( pBtnData->HasAfterwards() )
	{
		nRecursions++;

		if ( nRecursions < CApplication::GetInstance().Config().GetButtonCount() )
		{
			// REMARK: Recursion is possible!
			// The possible recursion is solved after execution 
			// reaches the number of configured buttons!
			this->ExecuteButton
				( CApplication::GetInstance().Config().FindAfterwardsButton
					( pBtnData->Afterwards() 
					) 
				);
		}
		else
		{
			nRecursions = 0;
		}
	}
}

bool CDialog::LaunchProcess
	( const TCHAR* szCommandLine
	, bool         bWait 
	)
{
	bool                success       = true; // Assume success!
	STARTUPINFO         siStartUpInfo = {0};
	PROCESS_INFORMATION piProcessInfo = {0};
	MSG                 sMessage      = {0};
	DWORD               nWaitResult   = WAIT_TIMEOUT;

	siStartUpInfo.cb          = sizeof(STARTUPINFO);
	siStartUpInfo.dwFlags     = STARTF_USESHOWWINDOW;
	siStartUpInfo.wShowWindow = SW_SHOWNORMAL;

	if ( ::CreateProcess
		( NULL
		, const_cast<TCHAR*>( szCommandLine )
		, NULL
		, NULL
		, FALSE
		, CREATE_DEFAULT_ERROR_MODE
		, NULL
		, NULL
		, &siStartUpInfo
		, &piProcessInfo 
		) )
	{
		if ( bWait )
		{
			CWaitCursorEx wait;

			this->m_bExecuting = true; // Lock waiting!

			while ( nWaitResult != WAIT_OBJECT_0 )
			{
				nWaitResult = ::MsgWaitForMultipleObjects
					( 1
					, &piProcessInfo.hProcess
					, FALSE
					, 1            // Ensure the thread switch.
					, QS_ALLEVENTS // Process WM_PAINT and other input stuff like WM_QUIT.
					); 

				switch ( nWaitResult ) 
				{
				case WAIT_OBJECT_0:
					{
						this->m_bExecuting = false; // Unlock waiting!
					}
					break;

				case ( WAIT_OBJECT_0 + 1 ):
					{
						// Process all available messages within the message queue.
						while ( ::PeekMessage( &sMessage, NULL, 0, 0, PM_REMOVE ) ) 
						{
							if ( sMessage.message == WM_QUIT )
							{
								// Close process handle because of exiting this function immediately.
								::CloseHandle( piProcessInfo.hProcess );

								// Close application window because of WM_QUIT message!
								this->OnClose();

								// Exit function with "success" because otherwise 
								// function ::ShellExecute() will be called!
								return success; 
							}

							// REMARK: Add additional message processing activities if needed!
							// If the main message loop does additional message processing - such 
							// as calling IsDialogMessage() for modeless dialogs - we will need to 
							// do those things here, too.
							::TranslateMessage( &sMessage );
							::DispatchMessage( &sMessage );
						}

						// Reset current cursor to the hourglass cursor.
						wait.UseCursor();
					}
					break;
				}
			}
		}

		// Close opened process handle.
		::CloseHandle( piProcessInfo.hProcess );
	}
	else
	{
		success = false;
	}

	return success;
}


